########################################################################
# OSM::Tree

=encoding utf8

=head1 NAME

OSM::Tree -- use OpenStreetMapData

=head1 SYNOPSIS

 #!/usr/bin/perl
 use strict;
 use warnings;
 use OSM::Tree;
 use Data::Dumper;

 my $osm=OSM::Tree->new('map2.osm');
 # or:
 # my $osmdata=' ... some osmxml ... ';
 # my $osm=OSM::Tree->new($osmdata);

 # get Version
 print "FILE VERSION: ".$osm->version()."\n";
 # get Generator
 print "FILE GENERATOR: ".$osm->generator()."\n";

 print "\n#############################################################\n";

 print "Nodes:".$osm->node_list_length()."\n";
 print "WAYS:".$osm->way_list_length()."\n";
 print "Relations:".$osm->relation_list_length()."\n";

 print "\n#############################################################\n";

 # print boundaty Box
 print "BOUNDARY BOX: (".join(', ',$osm->bbox()).")\n";

 my ($map_lan,$map_lon)=$osm->center_of_node_list($osm->node_list());
 print "Center of the Map: $map_lan,$map_lon\n";

 print "\n#############################################################\n";

 # get the fist way in list
 # print ID and all avaiable tags
 my @ways=$osm->way_list();
 print "WAY:\n";
 print "  ID: ".$ways[0]->id()."\n";
 my $tags=$ways[0]->tags();
 if($ways[0]->has_tags())
 {
   print "  TAGS:\n";
   my $tags=$ways[0]->tags();
   for my $tag ($tags->tag_list())
   {
     print $tag->name()."->".join(';',$tag->value_list())."\n";
   }
 }
 print "\n";
 # print the center of all ways:
 for my $way ($osm->way_list())
 {
   my ($lon,$lat)=$way->center();
   print "Cener of the Way:".$way->id()." is $lon,$lat\n";
 }

 print "\n#############################################################\n";

 # get all nodes who has a tag with name "name"
 for my $node ($osm->node_list())
 {
   if($node->tag('name'))
   {
     print "NODE ID: ".$node->id()."\n";
     print "  NAME :".$node->tag('name')->value()."\n";
     print "  KOORDINATE: ".join(' x ', $node->coordinate())."\n\n";
   }
 }

 print "\n#############################################################\n";

 # get the first node with a way
 # and print all nodes of this way
 for my $node ($osm->node_list())
 {
   my $way=$node->way();
   if($way)
   {
     print "NODE ".$node->id()." id part of WAY: ".$way->id()."\n";
     print "All NODES of this WAY are:\n";
     for my $node_id ($way->node_id_list())
     {
       my $node=$osm->node($node_id);
       print "  ID: ".$node->id()."\n";
       if($node->has_tags())
       {
         print "   TAGS:\n";
         my @tags=$node->tags()->tag_list();
         map{ $_=$_->name()." = ".$_->value() }@tags;
         print "    ".join("\n    ", @tags)."\n";
       }
     }
     last;
   }
 }


=head1 DESCRIPTION

Parse OpenStreetMap XML file

=head1 AUTHOR

ToPeG

=head1 USING

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::Polygon>

=head1 METHODS

=over 4

=cut

# CHANGELOG:
#
# 0.0.1 initial Release
# 0.1.0 Change interface
#  all methods who returns lists now named as ..._list
# 0.2.0 Change interface
#  tags and tag returns Objects intead lists and values
# 0.2.1 Extend interface
#  TagList now have the Method has_key and has_tag.
#  tag_list and key_list returns all existing keys given by a List
#  is no List ist set all keys/tags will be returned
# 0.2.2
#  OSM::Tree::BASE has now has_tag.
# 0.2.3
#  correction in __parse. Now all OSM Data will be parsed.
# 0.2.4
#  create Method OST::Tree::bbox and OST::Tree::Way::bbox
#  returns the boundary box of the OSM-Document/Way
# 0.2.5
#  TagSearch now optional caseinsensitive
# 0.2.6
#  way now has the method 'length'
#  that calulate the length op the way.
#  litte problem in way->center solved
# 0.2.7
#  correct wrong length calculation
#  correct error in parser
# 0.3.0
#  now two parser are aviable
#   1. Default using XML::Simpe (some Problems with Arrays)
#   2. Falback an pure-perl-Parser,
#      who has no externel dependencies
#  the third option in OSN::Tree->new alowes to select one
#  actualy "Falback" is the default,
#  because of some problems with XML::Simple
# 0.3.1
#  correct some minor Problems with the Falback Parser
#  now the Parser should work probably
#  solve the UTF-8 Problem
# 0.3.2
#  split the File to get it isier to handle
# 0.3.3
#  bbox now working korrect.
#  UTF-8 encodeing implemnetned in Falbackparser
# 0.3.4
#  bbox problem with multi defindes bounds solved
#  get the bbox arrount all bounds
# 0.3.5
#  added "has_all_keys" in TagList
# 0.4.0
#  corrected error in Relation,
#   not all member-ids are represented in the OSM-File!
#  change OSM::Tree::Tag and OSM::Tree::Taglist behavor
#  now also not existing Tags are returned as objekts, but they are empty
#  using "overload" to make the acesses easier.
#  i hope it will break nothing...
# 0.4.1
#  set links to the other pods
#  correct typos
# 0.4.2
#  correct wrong handling of tags in BASE.pm
# 0.4.3
#   add user_id, changeset to BASE.pm
# 0.4.4
#   add has_tags to BASE.pm
# 0.5.0
#   add OSM::Tree::XML::BigFile OSM::Tree::XML::BigFilePP
#   speed up OSM::Tree::XML::Default
# 0.5.1
#   add parse
# 0.5.2
#   corrct errors when Object is missing for an ID
# 0.5.3
#   add filter option to node_list way_list and relation_list
# 0.6.0
#   add to_xml for all Objects
#   now it's possible to generate OSM-XML from OSM::Tree.
# 0.7.0
#   add
#     node_add, way_add, relation_add, member_add,
#     node_del, way_del, relation_del, member_del,
#     node_id_add, way_id_add, relation_id_add, member_id_add,
#     node_id_del, way_id_del, relation_id_del, member_id_del,
#     tag_add tag_del
#   now it's possible to generate a complete OSM::Tree.
# 0.8.0
#   add OSM::Tree::Filter
#   add OSM::Tree::_add_del
#   change behavior of node_add, way_add, relation_add
#   change behavior of node_del, way_del, relation_del
#   edit POD
# 0.8.1
#   correct to_xml
# 0.9.0
#   create OSM::Tree::Polygon
#   add polygon to OSEM::Tree::Way
# 0.9.1
#   create OSM::Tree::Relation::kind()
#   modify OSM::Tree::BASE::relation_list
# 0.9.2
#   add force_delete show_deleted
# 0.9.3
#   add version
# 0.9.4
#   warning if Math::Polygon is not installed
#   but "polygon()" is used
# 0.9.5
#   add OSM::Tree::Way::node_add_after
#   add OSM::Tree::Way::node_add_bevor
#   add OSM::Tree::Way::node_add_append
# 0.9.6
#   OSM::Tree::Way::center() patched
#   an error occured if a Way had no nodes.


package OSM::Tree;
use strict;
use warnings;
use utf8;
use Carp;
use POSIX;

use base qw(OSM::Tree::_add_del);

use OSM::Tree::Node;
use OSM::Tree::Way;
use OSM::Tree::Relation;
#use OSM::Tree::XML::Default;
#use OSM::Tree::XML::Falback;
#use OSM::Tree::XML::BigFile;

our $VERSION='0.9.6';

my $CLASS_ERROR='';

=item new([XMLDATA|FILENAME],<PARSER>)

create a new Tree-Object

parser types:

Falback

=over 4

plain perl robust OSM-Parser

=back

Default

=over 4

using XML::Parser (fastest parser)

=back

BigFile (BigFilePP)

=over 4

using a BerkleyDB to store the values (<filename>.db)
slower than "Default" but mutch less memory usage

=back

"OSM::Tree::XML::Default" is the Default

returns an OSM::Tree Object

=cut
sub new
{
  my $class=shift;
  my $self={};

  $self->{TREE}={};
  $self->{bbox}=undef;
  $self->{PARSER}=undef;
  $self->{ERROR}='';
  $self->{IGNORE_DELETED}=\do{my $t=0};
  $self->{FORCE_DELETE}=\do{my $t=1};

  bless($self,$class);

  my $data=undef;

  # DATA?
  $data=shift() if(@_>0);

  # select parser
  if(@_>0)
  {
    my $parser=shift;
    unless($self->set_parser($parser))
    {
      $CLASS_ERROR=qq(Parser "$parser" not exists!);
      return undef;
    }
  }

  # no parser select...
  unless($self->{PARSER})
  {
    unless($self->set_parser('Default'))
    {
      unless($self->set_parser('Falback'))
      {
        $CLASS_ERROR=qq(Can't set any Parser!);
        return undef;
      }
    }
  }

  if($data)
  {
    # is it data or filename...
    unless($self->set_data($data))
    {
      $CLASS_ERROR='not an OSM Document';
      unless($self->set_file($data))
      {
        $CLASS_ERROR=qq(not an OSM File);
        return undef;
      }
    }
  }

  return $self;
}

# private constructor
# to create an Objecht
# that hold only a fiew Objects
# usefull for filtering
sub _emb_new
{
  my $class=shift;
  $class=ref($class) if(ref($class));
  my $self=shift;

  $self->{TREE}=shift;
  $self->{bbox}=undef;
  $self->{PARSER}=undef;
  $self->{IGNORE_DELETED}=\do{my $t=0};
  $self->{FORCE_DELETE}=\do{my $t=1};

  bless($self,$class);

  unless($self->{PARSER})
  {
    unless($self->set_parser('Default'))
    { $self->set_parser('Falback'); }
  }

  return $self;
}

=item parse([XMLDATA|FILENAME])

set new XML data

returns 1 on success else 0

=cut
sub parse
{
  my $self=shift;
  my $data=shift;
  if($data)
  {
    # is it data or filename...
    unless($self->set_data($data))
    {
      $self->{ERROR}='not an OSM Document';
      unless($self->set_file($data))
      {
        $self->{ERROR}=qq'not an OSM File';
        return 0;
      }
    }
  }
  return 1;
}

=item set_data(XMLDATA)

setup new XML data

returns 1 on success else 0

=cut
sub set_data
{
  my $self=shift;
  my $data=shift;
  $self->set_parser(shift);

  # is an osmfile?
  if(defined($data) && $data ne '' && $data=~/<osm\s*[^>]+>/s && $data=~m!</osm>!s)
  {
    # parse it
    my $tree=__parse($data,$self->{PARSER},'data');
    return $self->_set_tree($tree);
  }
  $self->{ERROR}='not an OSM Document';
  return 0;
}

=item set_file(FILENAME)

setup new XML data from a file

returns 1 on success else 0

=cut
sub set_file
{
  my $self=shift;
  my $file=shift;
  $self->set_parser(shift) if(@_);

  # is an osmfile?
  if($file && -f $file)
  {
    my $tree=__parse($file,$self->{PARSER},'file');
    return $self->_set_tree($tree);
  }
  $self->{ERROR}=qq(not an OSM file);
  return 0;
}

sub _set_tree
{
  my $self=shift;
  my $tree=shift;
  # sucessfull parsed?
  if(defined($tree) && ref($tree) eq 'HASH')
  {
    $self->{TREE}=$tree;

    # get bondary box if exists
    if(exists($self->{TREE}->{bound}))
    { $self->{bbox}=[split(',',$self->{TREE}->{bound}->{box})]; }
    elsif(exists($self->{TREE}->{bounds}))
    {
      my $box=$self->{TREE}->{bounds};
      my @bbox=(undef,undef,undef,undef);
      for my $bound (@{$self->{TREE}->{bounds}})
      {
        $bbox[0]=$bound->{minlat} unless defined($bbox[0]);
        $bbox[1]=$bound->{minlon} unless defined($bbox[1]);
        $bbox[2]=$bound->{maxlat} unless defined($bbox[2]);
        $bbox[3]=$bound->{maxlon} unless defined($bbox[3]);

        $bbox[0]=$bound->{minlat} if($bound->{minlat}<$bbox[0]);
        $bbox[2]=$bound->{maxlat} if($bound->{maxlat}>$bbox[2]);
        $bbox[1]=$bound->{minlon} if($bound->{minlon}<$bbox[1]);
        $bbox[3]=$bound->{maxlon} if($bound->{maxlon}>$bbox[3]);
      }

      if(defined($bbox[0]) && defined($bbox[1]) && defined($bbox[2]) && defined($bbox[3]))
      { $self->{bbox}=\@bbox; }
      else
      { $self->{bbox}=undef; }
    }
    else
    { $self->{bbox}=undef; }

    return 1;
  }
  return 0;
}

sub _id_list
{
  my $self=shift;
  my $name=shift;
  my $elms=$self->{TREE}->{$name};
  my @l=keys(%$elms);
  if(${$self->{IGNORE_DELETED}})
  {
    my @ret=();
    for(@l)
    { push(@ret,$_) if($elms->{$_}->{action} ne 'delete'); }
    return @ret;
  }
  return @l;
}
=item set_parser(type)

set new OSM Parser

returns 1 on success else 0

=cut
sub set_parser
{
  my $self=shift;
  my $parser=shift;
  if($parser)
  {
    for (qw(Falback Default BigFile BigFilePP))
    {
      if($parser eq $_ || $parser eq "OSM::Tree::XML::$_")
      {
        if($parser!~/^OSM::Tree::XML::/)
        { $parser="OSM::Tree::XML::$parser"; }
        eval("use $parser");
        unless($@)
        {
          $self->{PARSER}=$parser;
          return 1;
        }
      }
    }
  }
  return 0;
}

=item error()

get an OSM::Tree error

returns '' if there is no error

=cut
sub error
{
  my $self=shift;
  if($self && ref($self) eq __PACKAGE__)
  {
    my $e=$self->{ERROR};
    $self->{ERROR}='';
    return $e;
  }
  return $CLASS_ERROR;
}
#=======================================================================
sub __parse
{
  #print "\n__PARSE: ".join('; ',@_)."\n\n";
  my $data=shift;
  my $parser=shift || return undef;
  return undef unless($parser->can("parse"));
  return $parser->parse($data,@_);
}
#=======================================================================

=item version()

read OSM version

returns the version stored in OSM-XML

=cut
sub version
{ return shift->{TREE}->{version}; }

=item generator()

read name of XML-generator

returns the the OSM-generator stored in XML

=cut
sub generator
{ return shift->{TREE}->{generator}; }

=item bbox()

returns the BoundaryBox of the OSM Data

returns (lonmin, latmin, lonmax, latmax)

=cut
sub bbox
{
  my $self=shift;
  my $force=shift;

  # if no boundary box is defined create one
  if(!defined($self->{bbox}) || ref($self->{bbox}) ne 'ARRAY' || $force)
  {
    my ($lonmin,$latmin,$lonmax,$latmax)=(undef,undef,undef,undef);
    for my $nid ($self->node_id_list())
    {
      my $nref=$self->{TREE}->{node}->{$nid};
      my ($lon,$lat)=($nref->{lon},$nref->{lat});

      if(defined($lonmin))
      { $lonmin = $lon if($lonmin > $lon); }
      else
      { $lonmin = $lon; }

      if(defined($lonmax))
      { $lonmax = $lon if($lonmax < $lon); }
      else
      { $lonmax = $lon; }

      if(defined($latmin))
      { $latmin = $lat if($latmin > $lat); }
      else
      { $latmin = $lat; }

      if(defined($latmax))
      { $latmax = $lat if($latmax < $lat); }
      else
      { $latmax = $lat; }
    }
    $self->{bbox}=[$latmin,$lonmin,$latmax,$lonmax];
  }

  return @{$self->{bbox}};
}

=item show_deleted(<1/0>)

If it's set the nodes/ways/relation with "action='delete'" are listet as normal
returns the actual state

It's not set by default

=cut
sub show_deleted
{
  my $self=shift();
  ${$self->{IGNORE_DELETED}}=shift()?1:0 if(@_);
  return ${$self->{IGNORE_DELETED}}?1:0;
}

=item force_delete(<1/0>)

If it's set the nodes/ways/relation are deleted from the Stack.
If it's not set the tag "action='delete'" is set

It's set by default

returns the actual state

=cut
sub force_delete
{
  my $self=shift();
  ${$self->{FORCE_DELETE}}=shift()?1:0 if(@_);
  return ${$self->{FORCE_DELETE}}?1:0;
}


#=======================================================================

=item node(ID)

get node-object for an node-ID
see L<OSM::Tree::Node> for more

returns the an node-object or undef if there is no node with this ID

=cut
sub node
{
  my $self=shift;
  my $nodeid=shift;
  my $ret=OSM::Tree::Node->_new($self->{TREE},$nodeid,$self->{IGNORE_DELETED},$self->{FORCE_DELETE});
  return $ret if($ret);
  return wantarray?():undef;
}

=item node_add(<HASH>)

add node-object

keys are:

=over

=item * id

the unique node ID

=item * lon

longitude

=item * lat

latitude

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation dafault time()

=item * tags
hash of key/value pairs

=back

returns the an node-object or undef

see L<OSM::Tree::Node> and OSM-XML spec for more

=cut
sub node_add { &_node_add; }

=item node_del(NODE)

detete a node
returns 1 on success

=cut
sub node_del { &_node_del; }

=item node_id_del(ID)

detete a node
returns 1 on success

=cut
sub node_id_del { &_node_id_del; }

=item node_list(<CODE>)

get a node-object-list of all nodes
if an coderefence is given it will use as filter.

see L<OSM::Tree::Node> for more

returns a list of node-objects or an empty list if there are no nodes

=cut
sub node_list
{
  my $self=shift;
  my $filter=shift;
  my @list;
  if($filter)
  {
    if(ref($filter) eq 'CODE')
    {
      for($self->node_id_list())
      {
        my $node=$self->node($_);
        push(@list,$node) if($filter->($node))
      }
    }
    else
    { croak(q(Given filter is not a CodeRefence!)); }
  }
  else
  { @list=map{$self->node($_)}$self->node_id_list(); }
  return @list;
}

=item node_id_list()

get a list of all node-IDs

returns a list of node-IDs or an empty list if there are no nodes

=cut
sub node_id_list { return shift->_id_list('node'); }

=item node_list_length()

get the length the list of all nodes

returns the length of the node-list

=cut
sub node_list_length
{ return scalar(shift->node_id_list()); }

=item node_exists(ID)

returns true if the node-id exists

=cut
sub node_exists
{
  my $self=shift;
  my $nid=shift;
  return 0 unless(defined($nid));
  if(ref($nid) && ref($nid) eq 'OSM::Tree::Node')
  { $nid=$nid->id(); }
  return 0 if($nid=~/\D/);
  for my $id ($self->node_id_list())
  { return 1 if($nid==$id); }
  return 0;
}

#=======================================================================

=item way(ID)

get a way-object with this ID

see L<OSM::Tree::Way> for more

returns an way-object if this ID or undef if there is no id with this id

=cut
sub way
{
  my $self=shift;
  my $wayid=shift;
  my $ret=OSM::Tree::Way->_new($self->{TREE},$wayid,$self->{IGNORE_DELETED},$self->{FORCE_DELETE});
  return $ret if($ret);
  return wantarray?():undef;
}

=item way_add(<HASH>)

add way-object

keys are:

=over

=item * id

the unique node ID

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation Default: time()

=item * tags

hash of key/value pairs

=item * nodes

List of Node-IDs

=back

returns the an node-object or undef

see L<OSM::Tree::Way> and OSM-XML spec for more

=cut
sub way_add { &_way_add; }

=item way_del(WAY)

detete a node
returns 1 on success

=cut
sub way_del { &_way_del; }

=item way_id_del(ID)

detete a way
returns 1 on success

=cut
sub way_id_del { &_way_id_del; }

=item way_list(<CODE>)

get a list of all way-Objects
if an coderefence is given it will use as filter.

see L<OSM::Tree::Way> for more

returns a list of way-Objects
or an empty list if there are no ways

=cut
sub way_list
{
  my $self=shift;
  my $filter=shift;
  my @list;
  if($filter)
  {
    if(ref($filter) eq 'CODE')
    {
      for($self->way_id_list())
      {
        my $way=$self->way($_);
        push(@list,$way) if($filter->($way))
      }
    }
    else
    { croak(q(Given filter is not a CodeRefence!)); }
  }
  else
  { @list=map{$self->way($_)}$self->way_id_list(); }
  return @list;
}

=item way_id_list()

get a list of all way-IDs

returns a list of way-IDs
or an empty list if there are no ways

=cut
sub way_id_list
{ return shift->_id_list('way'); }

=item way_list_length()

get the length the list of all ways

returns the length of the way-list

=cut
sub way_list_length
{ return scalar(shift->way_id_list()); }

=item way_exists(ID)

returns true if the way-id exists

=cut
sub way_exists
{
  my $self=shift;
  my $nid=shift;
  return 0 unless(defined($nid));
  if(ref($nid) && ref($nid) eq 'OSM::Tree::Way')
  { $nid=$nid->id(); }
  return 0 if($nid=~/\D/);
  for my $id ($self->way_id_list())
  { return 1 if($nid==$id); }
  return 0;
}
#=======================================================================

=item relation(ID)

get an relation-object with this id

see L<OSM::Tree::Relation> for more

returns an relation.object
or undef if thre is no relation with this id

=cut
sub relation
{
  my $self=shift;
  my $relid=shift;
  my $ret=OSM::Tree::Relation->_new($self->{TREE},$relid,$self->{IGNORE_DELETED},$self->{FORCE_DELETE});
  return $ret if($ret);
  return wantarray?():undef;
}

=item relation_add(<HASH>)

add relation-object

keys are:

=over

=item * id

the unique node ID

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation Default: time()

=item * tags

hash of key/value pairs

=item * members

List of Hashes with ref type role see L<OSM::Tree::Member|OSM::Tree::Relation/OSM::Tree::Member> for more

=back

returns the an node-object or undef

see L<OSM::Tree::Way> and OSM-XML spec for more

=cut
sub relation_add { &_relation_add; }

=item relation_del(RELATION)

detete a relation
returns 1 on success

=cut
sub relation_del{ &_relation_del; }

=item relation_id_del(ID)

detete a relation
returns 1 on success;

=cut
sub relation_id_del { &_relation_id_del; }


=item relation_list(<CODE>)

get a list of all relation-Objects
if an coderefence is given it will use as filter.

see L<OSM::Tree::Relation> for more

returns a list of relation-Objects
or an empty list if there are no relations

=cut
sub relation_list
{
  my $self=shift;
  my $filter=shift;
  my @list;
  if($filter)
  {
    if(ref($filter) eq 'CODE')
    {
      for($self->relation_id_list())
      {
        my $relation=$self->relation($_);
        push(@list,$relation) if($filter->($relation))
      }
    }
    else
    { croak(q(Given filter is not a CodeRefence!)); }
  }
  else
  { @list=map{$self->relation($_)}$self->relation_id_list(); }
  return @list;
}

=item relation_id_list()

get a list of all relation-IDs

returns a list of relation-IDs
or an empty list if there are no relations

=cut
sub relation_id_list
{ return shift->_id_list('relation'); }

=item relation_list_length()

get the length the list of all relations

returns the length of the relation-list

=cut
sub relation_list_length
{ return scalar(shift->relation_id_list()); }

=item relation_exists(ID)

returns true if the relation-id exists

=cut
sub relation_exists
{
  my $self=shift;
  my $nid=shift;
  return 0 unless(defined($nid));
  if(ref($nid) && ref($nid) eq 'OSM::Tree::Relation')
  { $nid=$nid->id(); }
  return 0 if($nid=~/\D/);
  for my $id ($self->relation_id_list())
  { return 1 if($nid==$id); }
  return 0;
}

#=======================================================================

=item center_of_coordinate_list()

get the center of a list of kooardiantepairs

returns a longitude,latitude pair

Examples:

 my ($lon,$lat)= $osm->center_of_coordinate_list(
                                 [1,3],
                                 [12,5],
                                 [12,5.3221],
                                 [13,7]);
 my ($lon,$lat)= OSM::Tree::center_of_coordinate_list(
                                 [1,3],
                                 [12,5],
                                 [12,5.3221],
                                 [13,7]);

=cut
sub center_of_coordinate_list
{
  shift if(ref($_[0]) eq __PACKAGE__);
  my @list=@_;
  my $mx=0;
  my $my=0;
  for my $p (@list)
  {
    $mx+=$p->[0];
    $my+=$p->[1];
  }
  $mx/=@list;
  $my/=@list;
  return ($mx,$my);
}

=item center_of_node_list()

get the center of a list of Nodes oder Nodeids (can be mixed)

returns a longitude,latitude pair

Example:

 my ($lon,$lat)= $osm->center_of_coordinate_list(1,2,4,6,9,$nodea,$nodeb);

=cut
sub center_of_node_list
{
  my $self=shift;
  my @list=@_;
  my $mx=0;
  my $my=0;
  for my $p (@list)
  {
    $p=$self->node($p) unless(ref($p) eq __PACKAGE__.'::Node');
    return(undef,undef) unless(ref($p) eq __PACKAGE__.'::Node');
    $mx+=$p->longitude();
    $my+=$p->latitude();
  }
  $mx/=@list;
  $my/=@list;
  return ($mx,$my);
}

=item center_of_way()

get the center of a way or wayid

returns a longitude,latitude pair

Examples:

 my ($lon,$lat)= $osm->center_of_way($wayid);
 my ($lon,$lat)= $osm->center_of_way($wayref);

=cut
sub center_of_way
{
  my $self=shift;
  my $way=shift;
  # same as:
  #   $way=$self->way($way) unless($way->type() eq 'Way');
  # only faster and a litele bit more robust
  $way=$self->way($way) unless(ref($way) eq __PACKAGE__.'::Way');
  return $way->center() if($way);
  return (undef,undef);
}

=item length_of_way()

get the length of a way or wayid

returns a distance

Examples:

 my $length= $osm->length_of_way($wayid);
 my $length= $osm->length_of_way($wayref);

=cut
sub length_of_way
{
  my $self=shift;
  my $way=shift;
  $way=$self->way($way) unless(ref($way) eq __PACKAGE__.'::Way');
  return $way->length() if($way);
  return (-1);
}

=item to_xml()

convert OSM::Tree back to OSM-XML

returns the a String of the OSM-XML

=cut
sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $ref=$self->{TREE};
  my $origin=$ref->{origin} || 'OSM::Tree';
  my $ret=qq!$in<?xml version='1.0' encoding='UTF-8'?>\n!;
  $ret  .=qq!$in<osm version="0.6" generator="OSM::Tree $VERSION">\n!;
  my @box=$self->bbox();
  $ret  .=qq!$in  <bounds minlat="$box[0]" minlon="$box[1]" maxlat="$box[2]" maxlon="$box[3]" />\n!;
  $ret  .=$_->to_xml("$in  ") for(sort{$a->id() <=> $b->id()}$self->node_list());
  $ret  .=$_->to_xml("$in  ") for(sort{$a->id() <=> $b->id()}$self->way_list());
  $ret  .=$_->to_xml("$in  ") for(sort{$a->id() <=> $b->id()}$self->relation_list());
  $ret  .="$in</osm>\n";
  return $ret;
}

1;

=back

=cut

=head1 BUGS

None known.

Please send bugreports!

=head1 TODO

- distances beween Ways/Nodes

- crossing of ways

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree::Polygon>

L<OSM::Tree>

L<OSM::Filter>

=head1 AUTHOR

Tobias Grönhagen <osm@topeg.de>

=head1 VERSION

Version 0.9.6

=head1 COPYRIGHT

Copyright (c) Tobias Grönhagen. All rights reserved.

=head1 LICENSE

This program is free software; you can redistribute it
and/or modify it under the same terms and conditions as
Perl itself.

This means that you can, at your option, redistribute it and/or modify it under
either the terms the GNU Public License (GPL) version 2 or later, or under the
Perl Artistic License.

See http://dev.perl.org/licenses/

=head1 DISCLAIMER

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.

Use of this software in any way or in any form, source or binary,
is not allowed in any country which prohibits disclaimers of any
implied warranties of merchantability or fitness for a particular
purpose or any disclaimers of a similar nature.

IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION (INCLUDING, BUT NOT
LIMITED TO, LOST PROFITS) EVEN IF I HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE

=cut
