########################################################################
# OSM::Tree::XML::BigFilePP

# if no other parser is Avaiable
# this parser would be used

package OSM::Tree::XML::BigFilePP;
use strict;
use warnings;
use Encode;
use BerkeleyDB;
use IO::Scalar;

sub parse
{
  my $class=shift;
  my $data=shift;
  my $type=shift;

  my $fh=undef;
  my $db_name='temp.db';

  if($type eq 'file')
  {
    $db_name="$data.db";
    unlink($db_name) if(-f $db_name && -f $data && (stat($db_name))[9] < (stat($data))[9] );
    open($fh, '<', $data) // return undef;
  }
  else
  {
    $fh = new IO::Scalar \$data;
    unlink($db_name) if( -f $db_name );
  }

  my $db=__osm_parse($fh,$db_name);
  close($fh);

  my %hash;
  tie(%hash,__PACKAGE__, $db, 'data');
  return \%hash;
}

sub db
{
  my $self=shift;
  $self->{db}=shift if(@_ && ref($_[0]) && $_[0]->isa('BerkeleyDB::Common'));
  return $self->{db};
}

sub TIEHASH
{
  my $class=shift;
  my $db=shift;
  my $path=shift;

  my $self=bless({},$class);
  $self->{db}=$db;
  $self->{path}=$path || 'data';
  $self->{type}='HASH';

  my $list;
  $list='' if($self->db->db_get($path,$list)!=0);
  $self->{list}=[grep{$_}split("\0",$list)];
  $self->{pos}=0;

  #$list=~tr/\0/,/; print ($self->{path}." => $list\n");

  return $self;
}

sub FETCH
{
  my $self=shift;
  my $key=shift;
  my $val='';

  if($self->{path} eq 'data')
  {

    $self->db->db_get("data/$key",$val);

    if($key eq 'node')
    {
      my %data;
      tie(%data, __PACKAGE__, $self->db, 'data/node');
      return \%data;
    }
    elsif($key eq 'way')
    {
      my %data;
      tie(%data, __PACKAGE__, $self->db, 'data/way');
      return \%data;
    }
    elsif($key eq 'relation')
    {
      my %data;
      tie(%data, __PACKAGE__, $self->db, 'data/relation');
      return \%data;
    }
    elsif($key eq 'bounds')
    {
      my $count=$val;
      $val=[];
      if($count)
      {
        for my $pos (0..($count-1))
        {
          for my $k (qw(minlat minlon maxlat maxlon origin))
          {
            my $v;
            $self->db->db_get("data/bounds/$pos/$k",$v);
            $val->[$pos]->{$k}=$v || '';
          }
        }
      }
    }
    else
    { $self->db->db_get("data/$key",$val); }
  }
  elsif($self->{path} eq 'data/node')
  {
    if($self->db->db_get($self->{path}."/$key",$val)==0)
    {
      my @list=split("\0",$val);

      #$val=~tr/\0/,/; print ($self->{path}." => $val\n");

      $val={};
      for my $k (@list)
      {
        if($k eq 'tag')
        {
          $val->{$k}=$self->_fetch_tag($key);
          next;
        }

        my $v;
        if($self->db->db_get($self->{path}."/$key/$k",$v)==0)
        { $val->{$k}=$v; }
      }
    }
  }
  elsif($self->{path} eq 'data/way')
  {
    if($self->db->db_get($self->{path}."/$key",$val)==0)
    {
      my @list=split("\0",$val);

      #$val=~tr/\0/,/; print ($self->{path}." => $val\n");

      $val={};
      for my $k (@list)
      {
        if($k eq 'tag')
        {
          $val->{$k}=$self->_fetch_tag($key);
          next;
        }

        if($k eq 'nd')
        {
          my $v;
          if($self->db->db_get($self->{path}."/$key/nd",$v)==0)
          { $val->{$k}=[split("\0",$v)]; }
          next;
        }

        my $v;
        if($self->db->db_get($self->{path}."/$key/$k",$v)==0)
        { $val->{$k}=$v; }
      }
    }
  }
  elsif($self->{path} eq 'data/relation')
  {
    if($self->db->db_get($self->{path}."/$key",$val)==0)
    {
      my @list=split("\0",$val);

      #$val=~tr/\0/,/; print ($self->{path}." => $val\n");

      $val={};
      for my $k (@list)
      {

        if($k eq 'tag')
        {
          $val->{$k}=$self->_fetch_tag($key);
          next;
        }

        if($k eq 'member')
        {
          my $v;
          if($self->db->db_get($self->{path}."/$key/member",$v)==0)
          {
            for my $m (split("\0",$v))
            {
              my $mv;

              $mv='' if($self->db->db_get($self->{path}."/$key/member/$m/type",$mv)!=0);
              $val->{$k}->{$m}->{type}=$mv;

              $mv='' if($self->db->db_get($self->{path}."/$key/member/$m/role",$mv)!=0);
              $val->{$k}->{$m}->{role}=$mv;
            }
          }
          next;
        }

        my $v;
        if($self->db->db_get($self->{path}."/$key/$k",$v)==0)
        { $val->{$k}=$v; }
      }
    }
  }
  else
  {
    $self->db->db_get($self->{path}."/$key",$val);
  }

  return $val;
}

sub _fetch_tag
{
  my $self=shift;
  my $key=shift;
  my $ret={};
  my $val='';
  if($self->db->db_get($self->{path}."/$key/tag",$val)==0)
  {
    my @list=split("\0",$val);
    for my $k (@list)
    {
      my $v;
      if($self->db->db_get($self->{path}."/$key/tag/$k",$v)==0)
      { $ret->{$k}=$v; }
    }
  }

  return $ret;
}

sub STORE
{
  my $self=shift;
  my $key=shift;
  my $val=shift;
}

sub DELETE
{
  my $self=shift;
  my $key=shift;
}

sub CLEAR
{
  my $self=shift;
  $self->db->db_del($_) for(@{$self->{list}});
}

sub EXISTS
{
  my $self=shift;
  my $key=shift;
  my $val;
  return $self->db->db_get($self->{path}."/$key",$val)==0;
}

sub FIRSTKEY
{
  my $self=shift;
  $self->{pos}=0;
  return $self->NEXTKEY;
}

sub NEXTKEY
{
  my $self=shift;
  my $lastkey=shift;
  return $self->{list}->[$self->{pos}++];
}

sub SCALAR
{
  my $self=shift;
  return scalar($self);
}

sub DESTROY
{
  my $self=shift;
  $self->db->db_sync();
  #$self->db->db_close();
}

sub UNTIE
{
  my $self=shift;
  $self->db->db_sync();
  #$self->db->db_close();
}

sub dump
{
  my $self=shift;
  my ($k, $v) = ("", "");
  my $cursor = $self->db->db_cursor() ;
  while ($cursor->c_get($k, $v, DB_NEXT) == 0)
  { print "$k -> $v\n" }
}

# parse xml
sub __osm_parse
{
  my $fh=shift;
  my $db_file=shift;

  my $ret=-f $db_file;

  my $db=BerkeleyDB::Hash->new(-Filename => $db_file, -Flags=>DB_CREATE);
  $db->filter_fetch_value( sub { $_=Encode::decode("UTF-8", $_); });
  $db->filter_store_value( sub { $_=Encode::encode("UTF-8", $_); });

  return $db if($ret);

  my $inside;
  my %pkeys;

  my %all;

  while(my $line=<$fh>)
  {
    chomp($line);
    $line=__xml_cleanup($line);
    next unless($line);
    next if($line=~/^\s*$/s);
    #print "  --> $line\n";

    if($line=~m!<(/?\w+)\s*([^/>]*)\s*/?>!)
    {
      my $name=$1;
      my $opts=__split_opts($2);

      if(substr($name,0,1) eq '/')
      {
        substr($name,0,1,'');
        $inside=undef if($inside && $inside eq $name);
      }
      else
      {
        $pkeys{$name}++;

        if($name eq 'osm')
        {
          delete($pkeys{$name});
          while(my ($k,$v) = each(%$opts))
          {
            $db->db_put("data/$k",$v);
            $pkeys{$k}++;
          }
        }
        elsif($name eq 'bounds')
        {
          $all{"data/bounds"}=0 unless($all{"data/bounds"});

          while(my ($k,$v) = each(%$opts))
          { $db->db_put("data/bounds/".$all{"data/bounds"}."/$k",$v); }

          $all{"data/bounds"}++;
        }
        elsif($name eq 'bound')
        {
          delete($pkeys{'bound'});
          $pkeys{'bounds'}++;

          $all{"data/bounds"}=0 unless($all{"data/bounds"});

          my @box=split(",",$opts->{box});
          for my $k (qw(minlat minlon maxlat maxlon))
          { $db->db_put("data/bounds/".$all{"data/bounds"}."/$k",shift(@box)); }

          while(my ($k,$v) = each(%$opts))
          {
            next if($k eq 'box');
            $db->db_put("data/bounds/".$all{"data/bounds"}."/$k",$v);
          }

          $all{"data/bounds"}++;
        }
        elsif($name eq 'node')
        {
          my $id=$opts->{id};

          push(@{$all{"data/$name"}},$id);

          my $list="tag";
          while(my ($k,$v) = each(%$opts))
          {
            $list.="\0$k";
            $db->db_put("data/$name/$id/$k",$v);
          }
          $db->db_put("data/$name/$id",$list);

          $inside="data/$name/$id";
        }
        elsif($name eq 'way')
        {
          my $id=$opts->{id};

          push(@{$all{"data/$name"}},$id);

          my $list="nd\0tag";
          while(my ($k,$v) = each(%$opts))
          {
            $list.="\0$k";
            $db->db_put("data/$name/$id/$k",$v);
          }
          $db->db_put("data/$name/$id",$list);

          $inside="data/$name/$id";
        }
        elsif($name eq 'relation')
        {
          my $id=$opts->{id};

          push(@{$all{"data/$name"}},$id);

          my $list="member\0tag";
          while(my ($k,$v) = each(%$opts))
          {
            $list.="\0$k";
            $db->db_put("data/$name/$id/$k",$v);
          }
          $db->db_put("data/$name/$id",$list);

          $inside="data/$name/$id";
        }
        elsif($inside)
        {
          delete($pkeys{$name});
          my $p=$inside;

          if($name eq 'tag')
          {
            my $id=$opts->{k};
            my $v=$opts->{v};
            $db->db_put("$p/tag/$id",$v);

            push(@{$all{"$p/$name"}},$id);
          }
          elsif($name eq 'nd')
          {
            push(@{$all{"$p/$name"}},$opts->{ref});
          }
          elsif($name eq 'member')
          {
            my $id  =$opts->{ref};
            my $type=$opts->{type};
            my $role=$opts->{role};

            push(@{$all{"$p/$name"}},$id);

            $db->db_put("$p/member/$id/type",$type);
            $db->db_put("$p/member/$id/role",$role);
          }
          else
          { die("UNKOWN SUBTYPE $name ($line)\n"); }
        }
        #else
        #{ die("UNKOWN TYPE $name ($line)\n"); }

        #$inside=undef if($single);
      }
    }
  }

  $db->db_put("data",join("\0",sort keys(%pkeys)));

  while(my ($k,$v)=each(%all))
  {
    $v=join("\0",@$v) if(ref($v));
    $db->db_put($k,$v);
  }

  #warn join(",",keys(%pkeys))."\n";

  return $db;
}

sub __split_opts
{
  my $opts=shift || '';
  my $ref={};
  while($opts=~m!(\w+)\s*=\s*(["'])(.*?)(\2)!gcs)
  {
    $ref->{$1}=__xml_clean_values($3);
  }
  return $ref;
}

################
# remove all useless spaces an newlines
sub __xml_cleanup
{
  my $data=shift;
  # TODO:
  # CDATA harausholen, um es später in die Variable zu setzen...
  $data=~s/<!\[CDATA\[.+?\]\]>//gs;
  # kommentare fliegen raus
  $data=~s/<!--.+?-->//gs;
  # <? <! etc fliegt raus
  $data=~s!<[^/\w][^<>]+?\s*(?:\w+=(?:"[^"]*"|'[^']*')\s*)*\s*?>!!gs;
  # alle unnötigen leerzeichen fliegen raus
  $data=~s/\s+</</gs;
  $data=~s/>\s+/>/gs;
  $data=~s/\n\s+/\n/gs;
  return $data;
}

#########################
# list of XML-Characteres
my %xmlchars=(
              '&lt;'      => '<',
              '&gt;'      => '>',
              '&apos;'    => "'",
              '&quot;'    => '"',
              '&amp;'     => '&',
              '&apos;'    => "\x{27}",
              '&quot;'    => "\x{22}",
              '&nbsp;'    => "\x{A0}",
              '&iexcl;'   => "\x{A1}",
              '&cent;'    => "\x{A2}",
              '&pound;'   => "\x{A3}",
              '&curren;'  => "\x{A4}",
              '&yen;'     => "\x{A5}",
              '&brvbar;'  => "\x{A6}",
              '&sect;'    => "\x{A7}",
              '&uml;'     => "\x{A8}",
              '&copy;'    => "\x{A9}",
              '&ordf;'    => "\x{AA}",
              '&laquo;'   => "\x{AB}",
              '&not;'     => "\x{AC}",
              '&shy;'     => "\x{AD}",
              '&reg;'     => "\x{AE}",
              '&macr;'    => "\x{AF}",
              '&deg;'     => "\x{B0}",
              '&plusmn;'  => "\x{B1}",
              '&sup2;'    => "\x{B2}",
              '&sup3;'    => "\x{B3}",
              '&acute;'   => "\x{B4}",
              '&micro;'   => "\x{B5}",
              '&para;'    => "\x{B6}",
              '&middot;'  => "\x{B7}",
              '&cedil;'   => "\x{B8}",
              '&sup1;'    => "\x{B9}",
              '&ordm;'    => "\x{BA}",
              '&raquo;'   => "\x{BB}",
              '&frac14;'  => "\x{BC}",
              '&frac12;'  => "\x{BD}",
              '&frac34;'  => "\x{BE}",
              '&iquest;'  => "\x{BF}",
              '&Agrave;'  => "\x{C0}",
              '&Aacute;'  => "\x{C1}",
              '&Acirc;'   => "\x{C2}",
              '&Atilde;'  => "\x{C3}",
              '&Auml;'    => "\x{C4}",
              '&Aring;'   => "\x{C5}",
              '&AElig;'   => "\x{C6}",
              '&Ccedil;'  => "\x{C7}",
              '&Egrave;'  => "\x{C8}",
              '&Eacute;'  => "\x{C9}",
              '&Ecirc;'   => "\x{CA}",
              '&Euml;'    => "\x{CB}",
              '&Igrave;'  => "\x{CC}",
              '&Iacute;'  => "\x{CD}",
              '&Icirc;'   => "\x{CE}",
              '&Iuml;'    => "\x{CF}",
              '&ETH;'     => "\x{D0}",
              '&Ntilde;'  => "\x{D1}",
              '&Ograve;'  => "\x{D2}",
              '&Oacute;'  => "\x{D3}",
              '&Ocirc;'   => "\x{D4}",
              '&Otilde;'  => "\x{D5}",
              '&Ouml;'    => "\x{D6}",
              '&times;'   => "\x{D7}",
              '&Oslash;'  => "\x{D8}",
              '&Ugrave;'  => "\x{D9}",
              '&Uacute;'  => "\x{DA}",
              '&Ucirc;'   => "\x{DB}",
              '&Uuml;'    => "\x{DC}",
              '&Yacute;'  => "\x{DD}",
              '&THORN;'   => "\x{DE}",
              '&szlig;'   => "\x{DF}",
              '&agrave;'  => "\x{E0}",
              '&aacute;'  => "\x{E1}",
              '&acirc;'   => "\x{E2}",
              '&atilde;'  => "\x{E3}",
              '&auml;'    => "\x{E4}",
              '&aring;'   => "\x{E5}",
              '&aelig;'   => "\x{E6}",
              '&ccedil;'  => "\x{E7}",
              '&egrave;'  => "\x{E8}",
              '&eacute;'  => "\x{E9}",
              '&ecirc;'   => "\x{EA}",
              '&euml;'    => "\x{EB}",
              '&igrave;'  => "\x{EC}",
              '&iacute;'  => "\x{ED}",
              '&icirc;'   => "\x{EE}",
              '&iuml;'    => "\x{EF}",
              '&eth;'     => "\x{F0}",
              '&ntilde;'  => "\x{F1}",
              '&ograve;'  => "\x{F2}",
              '&oacute;'  => "\x{F3}",
              '&ocirc;'   => "\x{F4}",
              '&otilde;'  => "\x{F5}",
              '&ouml;'    => "\x{F6}",
              '&divide;'  => "\x{F7}",
              '&oslash;'  => "\x{F8}",
              '&ugrave;'  => "\x{F9}",
              '&uacute;'  => "\x{FA}",
              '&ucirc;'   => "\x{FB}",
              '&uuml;'    => "\x{FC}",
              '&yacute;'  => "\x{FD}",
              '&thorn;'   => "\x{FE}",
              '&yuml;'    => "\x{FF}",
              '&Alpha;'   => "\x{391}",
              '&alpha;'   => "\x{3B1}",
              '&Beta;'    => "\x{392}",
              '&beta;'    => "\x{3B2}",
              '&Gamma;'   => "\x{393}",
              '&gamma;'   => "\x{3B3}",
              '&Delta;'   => "\x{394}",
              '&delta;'   => "\x{3B4}",
              '&Epsilon;' => "\x{395}",
              '&epsilon;' => "\x{3B5}",
              '&Zeta;'    => "\x{396}",
              '&zeta;'    => "\x{3B6}",
              '&Eta;'     => "\x{397}",
              '&eta;'     => "\x{3B7}",
              '&Theta;'   => "\x{398}",
              '&theta;'   => "\x{3B8}",
              '&Iota;'    => "\x{399}",
              '&iota;'    => "\x{3B9}",
              '&Kappa;'   => "\x{39A}",
              '&kappa;'   => "\x{3BA}",
              '&Lambda;'  => "\x{39B}",
              '&lambda;'  => "\x{3BB}",
              '&Mu;'      => "\x{39C}",
              '&mu;'      => "\x{3BC}",
              '&Nu;'      => "\x{39D}",
              '&nu;'      => "\x{3BD}",
              '&Xi;'      => "\x{39E}",
              '&xi;'      => "\x{3BE}",
              '&Omicron;' => "\x{39F}",
              '&omicron;' => "\x{3BF}",
              '&Pi;'      => "\x{3A0}",
              '&pi;'      => "\x{3C0}",
              '&Rho;'     => "\x{3A1}",
              '&rho;'     => "\x{3C1}",
              '&Sigma;'   => "\x{3A3}",
              '&sigmaf;'  => "\x{3C2}",
              '&sigma;'   => "\x{3C3}",
              '&Tau;'     => "\x{3A4}",
              '&tau;'     => "\x{3C4}",
              '&Upsilon;' => "\x{3A5}",
              '&upsilon;' => "\x{3C5}",
              '&Phi;'     => "\x{3A6}",
              '&phi;'     => "\x{3C6}",
              '&Chi;'     => "\x{3A7}",
              '&chi;'     => "\x{3C7}",
              '&Psi;'     => "\x{3A8}",
              '&psi;'     => "\x{3C8}",
              '&Omega;'   => "\x{3A9}",
              '&omega;'   => "\x{3C9}",
              '&thetasym;'=> "\x{3D1}",
              '&upsih;'   => "\x{3D2}",
              '&piv;'     => "\x{3D6}",
              '&forall;'  => "\x{2200}",
              '&part;'    => "\x{2202}",
              '&exist;'   => "\x{2203}",
              '&empty;'   => "\x{2205}",
              '&nabla;'   => "\x{2207}",
              '&isin;'    => "\x{2208}",
              '&notin;'   => "\x{2209}",
              '&ni;'      => "\x{220B}",
              '&prod;'    => "\x{220F}",
              '&sum;'     => "\x{2211}",
              '&minus;'   => "\x{2212}",
              '&lowast;'  => "\x{2217}",
              '&radic;'   => "\x{221A}",
              '&prop;'    => "\x{221D}",
              '&infin;'   => "\x{221E}",
              '&ang;'     => "\x{2220}",
              '&and;'     => "\x{2227}",
              '&or;'      => "\x{2228}",
              '&cap;'     => "\x{2229}",
              '&cup;'     => "\x{222A}",
              '&int;'     => "\x{222B}",
              '&there4;'  => "\x{2234}",
              '&sim;'     => "\x{223C}",
              '&cong;'    => "\x{2245}",
              '&asymp;'   => "\x{2248}",
              '&ne;'      => "\x{2260}",
              '&equiv;'   => "\x{2261}",
              '&le;'      => "\x{2264}",
              '&ge;'      => "\x{2265}",
              '&sub;'     => "\x{2282}",
              '&sup;'     => "\x{2283}",
              '&nsub;'    => "\x{2284}",
              '&sube;'    => "\x{2286}",
              '&supe;'    => "\x{2287}",
              '&oplus;'   => "\x{2295}",
              '&otimes;'  => "\x{2297}",
              '&perp;'    => "\x{22A5}",
              '&sdot;'    => "\x{22C5}",
              '&loz;'     => "\x{25CA}",
              '&lceil;'   => "\x{2308}",
              '&rceil;'   => "\x{2309}",
              '&lfloor;'  => "\x{230A}",
              '&rfloor;'  => "\x{230B}",
              '&lang;'    => "\x{2329}",
              '&rang;'    => "\x{232A}",
              '&larr;'    => "\x{2190}",
              '&uarr;'    => "\x{2191}",
              '&rarr;'    => "\x{2192}",
              '&darr;'    => "\x{2193}",
              '&harr;'    => "\x{2194}",
              '&crarr;'   => "\x{21B5}",
              '&lArr;'    => "\x{21D0}",
              '&uArr;'    => "\x{21D1}",
              '&rArr;'    => "\x{21D2}",
              '&dArr;'    => "\x{21D3}",
              '&hArr;'    => "\x{21D4}",
              '&bull;'    => "\x{2022}",
              '&prime;'   => "\x{2032}",
              '&Prime;'   => "\x{2033}",
              '&oline;'   => "\x{203E}",
              '&frasl;'   => "\x{2044}",
              '&weierp;'  => "\x{2118}",
              '&image;'   => "\x{2111}",
              '&real;'    => "\x{211C}",
              '&trade;'   => "\x{2122}",
              '&euro;'    => "\x{20AC}",
              '&alefsym;' => "\x{2135}",
              '&spades;'  => "\x{2660}",
              '&clubs;'   => "\x{2663}",
              '&hearts;'  => "\x{2665}",
              '&diams;'   => "\x{2666}",
              '&OElig;'   => "\x{152}",
              '&oelig;'   => "\x{153}",
              '&Scaron;'  => "\x{160}",
              '&scaron;'  => "\x{161}",
              '&Yuml;'    => "\x{178}",
              '&fnof;'    => "\x{192}",
              '&ensp;'    => "\x{2002}",
              '&emsp;'    => "\x{2003}",
              '&thinsp;'  => "\x{2009}",
              '&zwnj;'    => "\x{200C}",
              '&zwj;'     => "\x{200D}",
              '&lrm;'     => "\x{200E}",
              '&rlm;'     => "\x{200F}",
              '&ndash;'   => "\x{2013}",
              '&mdash;'   => "\x{2014}",
              '&lsquo;'   => "\x{2018}",
              '&rsquo;'   => "\x{2019}",
              '&ldquo;'   => "\x{201C}",
              '&rdquo;'   => "\x{201D}",
              '&dagger;'  => "\x{2020}",
              '&Dagger;'  => "\x{2021}",
              '&hellip;'  => "\x{2026}",
              '&permil;'  => "\x{2030}",
              '&lsaquo;'  => "\x{2039}",
              '&rsaquo;'  => "\x{203A}",
              '&circ;'    => "\x{2C6}",
              '&tilde;'   => "\x{2DC}",
              );

###############
# convert XML-Escaped Chars
sub __xml_clean_values
{
  my $str=shift;
  $str=~s/(&\w+;)/exists($xmlchars{$1}) ? $xmlchars{$1} : int($1)>0 ? chr($1) : $1/gse;
  $str=Encode::decode('UTF-8', $str);
  return $str;
}

1;
