########################################################################
# OSM::Tree::XML::Falback

# if no other parser is Avaiable
# this parser would be used

package OSM::Tree::XML::Falback;
use strict;
use warnings;
use Encode;
#use Data::Dumper;

sub parse
{
  #print "OSM::Tree::XML::Falback: @_\n";
  my $class=shift;
  my $data=shift;
  my $type=shift || 'data';
  $data=__slurp($data) if($type eq 'file');
  my $tree=__xml_parse($data);
  #print Dumper($tree);
  return __osm_transform($tree);
}

sub __slurp
{
  my $file=shift;
  return eval{local(@ARGV,$/)=($file); <>; };
}

# transforn the xml-data to an usable OSM-Tree-Hash
sub __osm_transform
{
  my $tree=shift;
  $tree=$tree->{osm} if(exists($tree->{osm}));

  __osm_clean_bounds($tree);

  # clean nodes
  for my $type (qw(node way relation))
  { __osm_base_clean($tree,$type); }

  for my $way (values(%{$tree->{way}}))
  { __osm_nd_clean($way); }

  for my $relation (values(%{$tree->{relation}}))
  { __osm_member_clean($relation) }

  return $tree;
}

sub __osm_clean_bounds
{
  my $tree=shift;
  if(exists($tree->{bounds}))
  {
    if(ref($tree->{bounds}) eq 'HASH')
    { $tree->{bounds}=[$tree->{bounds}]; }
    elsif(ref($tree->{bounds}) ne 'ARRAY')
    { delete($tree->{bounds}) }
  }
}

# clanup Node Way Relelation
sub __osm_base_clean
{
  my $tree=shift;
  my $type=shift;
  return $tree unless($type);
  if(exists($tree->{$type}))
  {
    if(ref($tree->{$type}) eq 'ARRAY')
    {
      my @list=@{$tree->{$type}};
      $tree->{$type}={};
      for my $node (@list)
      {
        if(exists($node->{id}))
        {
          __osm_obj_clean($node);
          __osm_tag_clean($node);
          $tree->{$type}->{$node->{id}}=$node;
        }
      }
    }
    elsif(ref($tree->{$type}) eq 'HASH')
    {
      my $node=$tree->{$type};
      if(exists($node->{id}))
      {
        __osm_tag_clean($node);
        $tree->{$type}={$node->{id}=>$node};
      }
      else
      { $tree->{$type}={}; }
    }
    else
    { $tree->{$type}={}; }
  }
  else
  { $tree->{$type}={}; }
  return $tree;
}

#clean the way/node/relation
sub __osm_obj_clean
{
  my $obj=shift;
  if(ref($obj) eq 'HASH')
  {
    for my $val (values(%$obj))
    {
      if(ref($val) eq 'HASH' && exists($val->{content}))
      { $val=$val->{content}; }
      elsif(ref($val) eq 'HASH' && !keys(%$val))
      { $val=''; }
    }
  }
}

# cleanup Tags
sub __osm_tag_clean
{
  my $obj=shift;
  if(ref($obj) eq 'HASH')
  {
    if(exists($obj->{tag}))
    {
      if(ref($obj->{tag}) eq 'ARRAY')
      {
        my @list=@{$obj->{tag}};
        $obj->{tag}={};
        for my $tag (@list)
        {
          if(ref($tag) eq 'HASH')
          {
            if(ref($tag->{k}) eq 'HASH')
            {$tag->{k}=$tag->{k}->{content}}
            if(ref($tag->{v}) eq 'HASH')
            {$tag->{v}=$tag->{v}->{content}}
            if(defined($tag->{k}) && defined($tag->{v}))
            {
              $obj->{tag}->{$tag->{k}}=$tag->{v};
            }
          }
        }
      }
      elsif( ref($obj->{tag}) eq 'HASH' && defined($obj->{tag}->{k}) && defined($obj->{tag}->{v}) )
      { $obj->{tag}={ $obj->{tag}->{k} => $obj->{tag}->{v} }; }
      else
      { $obj->{tag}={}; }
    }
    else
    { $obj->{tag}={}; }
  }
}

# cleanup way->nd
sub __osm_nd_clean
{
  my $obj=shift;
  if(ref($obj) eq 'HASH')
  {
    if(exists($obj->{nd}))
    {
      if(ref($obj->{nd}) eq 'ARRAY')
      {
        my @list=@{$obj->{nd}};
        $obj->{nd}=[];
        for my $ref (@list)
        {
          if(ref($ref) eq 'HASH' && defined($ref->{ref}))
          { push(@{$obj->{nd}},$ref->{ref}); }
        }
      }
      elsif(ref($obj->{nd}) eq 'HASH')
      {
        if(exists($obj->{nd}->{ref}))
        { $obj->{nd}=[$obj->{nd}->{ref}]; }
        else
        { $obj->{nd}=[]; }
      }
      else
      { $obj->{nd}=[]; }
    }
    else
    { $obj->{nd}=[]; }
  }
}

# clanup relation->member
sub __osm_member_clean
{
  my $obj=shift;
  if(ref($obj) eq 'HASH')
  {
    if(exists($obj->{member}))
    {
      if(ref($obj->{member}) eq 'ARRAY')
      {
        my @list=@{$obj->{member}};
        $obj->{member}={};
        for my $member (@list)
        {
          if(ref($member) eq 'HASH' && defined($member->{ref}))
          { $obj->{member}->{ $member->{ref} }=$member; }
        }
      }
      elsif(ref($obj->{member}) eq 'HASH' && defined($obj->{member}->{ref}))
      { $obj->{member}={$obj->{member}->{ref}=>$obj->{member}}; }
      else
      { $obj->{member}={}; }
    }
    else
    { $obj->{member}={}; }
  }
}

# parse xml
sub __xml_parse
{
  my $data=shift || '';
  $data=__xml_cleanup($data);
  my $tree={};
  my @deep=({name=>'',ref=>$tree});
  unless($data=~m!</?[\w\-_]+\s*(?:[\w\-_]+\s*=\s*(?:"[^"]*"|'[^']*')\s*)*\s*?/?>!s)
  { return {content=>$data}; }

  while($data=~m!<(/?[\w\-_]+)\s*((?:[\w\-_]+\s*=\s*(?:"[^"]*"|'[^']*')\s*)*)\s*?(/?)>([^<>]*)!gcs)
  {
    my $name=$1;
    my $opts=$2;
    my $single=$3;
    my $text=__xml_clean_values($4);

    if(substr($name,0,1) eq '/')
    {
      substr($name,0,1,'');
      return $tree unless(@deep>1);
      while(my $p=shift(@deep))
      { last if($p->{name} eq $name); }
      return $tree unless(@deep);
      $single=1;
    }
    else
    {
      my $ref=$deep[0]->{ref};

      if(ref($ref) eq 'ARRAY')
      {
        unshift(@deep,{name=>$name,ref=>{}});
        push(@$ref,$deep[0]->{ref});
      }
      elsif(exists($ref->{$name}))
      {
        unshift(@deep,{name=>$name,ref=>{}});
        $ref->{$name}=[$ref->{$name}] if(ref($ref->{$name}) ne 'ARRAY');
        push(@{$ref->{$name}},$deep[0]->{ref});
      }
      else
      {
        unshift(@deep,{name=>$name,ref=>{}});
        $ref->{$name}=$deep[0]->{ref};
      }

      $ref=$deep[0]->{ref};

      if(!$single and defined($text) and $text ne '')
      {
        if(ref($ref) eq 'ARRAY')
        { push(@{$ref},{content=>$text}); }
        else
        { $ref->{content}=$text; }
      }

      if($opts)
      {
        while($opts=~m!(\w+)\s*=\s*(["'])(.*?)(\2)!gcs)
        {
          $ref->{$1}=__xml_clean_values($3);
        }
      }

      if($single)
      {
        $deep[0]->{ref}->{content}='' if(ref($deep[0]->{ref}) eq 'HASH' && keys(%{$deep[0]->{ref}})==0);
        shift(@deep) if(ref($deep[0]->{ref}) ne 'ARRAY');
        $ref=$deep[0]->{ref};
      }
    }

    if($text && $single)
    {
      my $r=undef;
      if(ref($deep[0]->{ref}) eq 'HASH')
      { $r=$deep[0]->{ref}; }
      elsif(ref($deep[0]->{ref}) eq 'ARRAY' && exists($deep[0]->{ref}->[-1]))
      { $r=$deep[0]->{ref}->[-1]; }

      if($r)
      {
        if(exists($r->{content}))
        {
          if(ref($r->{content}) eq 'ARRAY')
          { push(@{$r->{content}},$text); }
          else
          { $r->{content}=[$r->{content},$text]; }
        }
        else
        { $r->{content}=$text; }
      }
    }
  }

  return $tree;
}

################
# remove all useless spaces an newlines
sub __xml_cleanup
{
  my $data=shift;
  # TODO:
  # CDATA harausholen, um es später in die Variable zu setzen...
  $data=~s/<!\[CDATA\[.+?\]\]>//gs;
  # kommentare fliegen raus
  $data=~s/<!--.+?-->//gs;
  # <? <! etc fliegt raus
  $data=~s!<[^/\w][^<>]+?\s*(?:\w+=(?:"[^"]*"|'[^']*')\s*)*\s*?>!!gs;
  # alle unnötigen leerzeichen fliegen raus
  $data=~s/\s+</</gs;
  $data=~s/>\s+/>/gs;
  $data=~s/\n\s+/\n/gs;
  return $data;
}

#########################
# list of XML-Characteres
my %xmlchars=(
              '&lt;'      => '<',
              '&gt;'      => '>',
              '&apos;'    => "'",
              '&quot;'    => '"',
              '&amp;'     => '&',
              '&apos;'    => "\x{27}",
              '&quot;'    => "\x{22}",
              '&nbsp;'    => "\x{A0}",
              '&iexcl;'   => "\x{A1}",
              '&cent;'    => "\x{A2}",
              '&pound;'   => "\x{A3}",
              '&curren;'  => "\x{A4}",
              '&yen;'     => "\x{A5}",
              '&brvbar;'  => "\x{A6}",
              '&sect;'    => "\x{A7}",
              '&uml;'     => "\x{A8}",
              '&copy;'    => "\x{A9}",
              '&ordf;'    => "\x{AA}",
              '&laquo;'   => "\x{AB}",
              '&not;'     => "\x{AC}",
              '&shy;'     => "\x{AD}",
              '&reg;'     => "\x{AE}",
              '&macr;'    => "\x{AF}",
              '&deg;'     => "\x{B0}",
              '&plusmn;'  => "\x{B1}",
              '&sup2;'    => "\x{B2}",
              '&sup3;'    => "\x{B3}",
              '&acute;'   => "\x{B4}",
              '&micro;'   => "\x{B5}",
              '&para;'    => "\x{B6}",
              '&middot;'  => "\x{B7}",
              '&cedil;'   => "\x{B8}",
              '&sup1;'    => "\x{B9}",
              '&ordm;'    => "\x{BA}",
              '&raquo;'   => "\x{BB}",
              '&frac14;'  => "\x{BC}",
              '&frac12;'  => "\x{BD}",
              '&frac34;'  => "\x{BE}",
              '&iquest;'  => "\x{BF}",
              '&Agrave;'  => "\x{C0}",
              '&Aacute;'  => "\x{C1}",
              '&Acirc;'   => "\x{C2}",
              '&Atilde;'  => "\x{C3}",
              '&Auml;'    => "\x{C4}",
              '&Aring;'   => "\x{C5}",
              '&AElig;'   => "\x{C6}",
              '&Ccedil;'  => "\x{C7}",
              '&Egrave;'  => "\x{C8}",
              '&Eacute;'  => "\x{C9}",
              '&Ecirc;'   => "\x{CA}",
              '&Euml;'    => "\x{CB}",
              '&Igrave;'  => "\x{CC}",
              '&Iacute;'  => "\x{CD}",
              '&Icirc;'   => "\x{CE}",
              '&Iuml;'    => "\x{CF}",
              '&ETH;'     => "\x{D0}",
              '&Ntilde;'  => "\x{D1}",
              '&Ograve;'  => "\x{D2}",
              '&Oacute;'  => "\x{D3}",
              '&Ocirc;'   => "\x{D4}",
              '&Otilde;'  => "\x{D5}",
              '&Ouml;'    => "\x{D6}",
              '&times;'   => "\x{D7}",
              '&Oslash;'  => "\x{D8}",
              '&Ugrave;'  => "\x{D9}",
              '&Uacute;'  => "\x{DA}",
              '&Ucirc;'   => "\x{DB}",
              '&Uuml;'    => "\x{DC}",
              '&Yacute;'  => "\x{DD}",
              '&THORN;'   => "\x{DE}",
              '&szlig;'   => "\x{DF}",
              '&agrave;'  => "\x{E0}",
              '&aacute;'  => "\x{E1}",
              '&acirc;'   => "\x{E2}",
              '&atilde;'  => "\x{E3}",
              '&auml;'    => "\x{E4}",
              '&aring;'   => "\x{E5}",
              '&aelig;'   => "\x{E6}",
              '&ccedil;'  => "\x{E7}",
              '&egrave;'  => "\x{E8}",
              '&eacute;'  => "\x{E9}",
              '&ecirc;'   => "\x{EA}",
              '&euml;'    => "\x{EB}",
              '&igrave;'  => "\x{EC}",
              '&iacute;'  => "\x{ED}",
              '&icirc;'   => "\x{EE}",
              '&iuml;'    => "\x{EF}",
              '&eth;'     => "\x{F0}",
              '&ntilde;'  => "\x{F1}",
              '&ograve;'  => "\x{F2}",
              '&oacute;'  => "\x{F3}",
              '&ocirc;'   => "\x{F4}",
              '&otilde;'  => "\x{F5}",
              '&ouml;'    => "\x{F6}",
              '&divide;'  => "\x{F7}",
              '&oslash;'  => "\x{F8}",
              '&ugrave;'  => "\x{F9}",
              '&uacute;'  => "\x{FA}",
              '&ucirc;'   => "\x{FB}",
              '&uuml;'    => "\x{FC}",
              '&yacute;'  => "\x{FD}",
              '&thorn;'   => "\x{FE}",
              '&yuml;'    => "\x{FF}",
              '&Alpha;'   => "\x{391}",
              '&alpha;'   => "\x{3B1}",
              '&Beta;'    => "\x{392}",
              '&beta;'    => "\x{3B2}",
              '&Gamma;'   => "\x{393}",
              '&gamma;'   => "\x{3B3}",
              '&Delta;'   => "\x{394}",
              '&delta;'   => "\x{3B4}",
              '&Epsilon;' => "\x{395}",
              '&epsilon;' => "\x{3B5}",
              '&Zeta;'    => "\x{396}",
              '&zeta;'    => "\x{3B6}",
              '&Eta;'     => "\x{397}",
              '&eta;'     => "\x{3B7}",
              '&Theta;'   => "\x{398}",
              '&theta;'   => "\x{3B8}",
              '&Iota;'    => "\x{399}",
              '&iota;'    => "\x{3B9}",
              '&Kappa;'   => "\x{39A}",
              '&kappa;'   => "\x{3BA}",
              '&Lambda;'  => "\x{39B}",
              '&lambda;'  => "\x{3BB}",
              '&Mu;'      => "\x{39C}",
              '&mu;'      => "\x{3BC}",
              '&Nu;'      => "\x{39D}",
              '&nu;'      => "\x{3BD}",
              '&Xi;'      => "\x{39E}",
              '&xi;'      => "\x{3BE}",
              '&Omicron;' => "\x{39F}",
              '&omicron;' => "\x{3BF}",
              '&Pi;'      => "\x{3A0}",
              '&pi;'      => "\x{3C0}",
              '&Rho;'     => "\x{3A1}",
              '&rho;'     => "\x{3C1}",
              '&Sigma;'   => "\x{3A3}",
              '&sigmaf;'  => "\x{3C2}",
              '&sigma;'   => "\x{3C3}",
              '&Tau;'     => "\x{3A4}",
              '&tau;'     => "\x{3C4}",
              '&Upsilon;' => "\x{3A5}",
              '&upsilon;' => "\x{3C5}",
              '&Phi;'     => "\x{3A6}",
              '&phi;'     => "\x{3C6}",
              '&Chi;'     => "\x{3A7}",
              '&chi;'     => "\x{3C7}",
              '&Psi;'     => "\x{3A8}",
              '&psi;'     => "\x{3C8}",
              '&Omega;'   => "\x{3A9}",
              '&omega;'   => "\x{3C9}",
              '&thetasym;'=> "\x{3D1}",
              '&upsih;'   => "\x{3D2}",
              '&piv;'     => "\x{3D6}",
              '&forall;'  => "\x{2200}",
              '&part;'    => "\x{2202}",
              '&exist;'   => "\x{2203}",
              '&empty;'   => "\x{2205}",
              '&nabla;'   => "\x{2207}",
              '&isin;'    => "\x{2208}",
              '&notin;'   => "\x{2209}",
              '&ni;'      => "\x{220B}",
              '&prod;'    => "\x{220F}",
              '&sum;'     => "\x{2211}",
              '&minus;'   => "\x{2212}",
              '&lowast;'  => "\x{2217}",
              '&radic;'   => "\x{221A}",
              '&prop;'    => "\x{221D}",
              '&infin;'   => "\x{221E}",
              '&ang;'     => "\x{2220}",
              '&and;'     => "\x{2227}",
              '&or;'      => "\x{2228}",
              '&cap;'     => "\x{2229}",
              '&cup;'     => "\x{222A}",
              '&int;'     => "\x{222B}",
              '&there4;'  => "\x{2234}",
              '&sim;'     => "\x{223C}",
              '&cong;'    => "\x{2245}",
              '&asymp;'   => "\x{2248}",
              '&ne;'      => "\x{2260}",
              '&equiv;'   => "\x{2261}",
              '&le;'      => "\x{2264}",
              '&ge;'      => "\x{2265}",
              '&sub;'     => "\x{2282}",
              '&sup;'     => "\x{2283}",
              '&nsub;'    => "\x{2284}",
              '&sube;'    => "\x{2286}",
              '&supe;'    => "\x{2287}",
              '&oplus;'   => "\x{2295}",
              '&otimes;'  => "\x{2297}",
              '&perp;'    => "\x{22A5}",
              '&sdot;'    => "\x{22C5}",
              '&loz;'     => "\x{25CA}",
              '&lceil;'   => "\x{2308}",
              '&rceil;'   => "\x{2309}",
              '&lfloor;'  => "\x{230A}",
              '&rfloor;'  => "\x{230B}",
              '&lang;'    => "\x{2329}",
              '&rang;'    => "\x{232A}",
              '&larr;'    => "\x{2190}",
              '&uarr;'    => "\x{2191}",
              '&rarr;'    => "\x{2192}",
              '&darr;'    => "\x{2193}",
              '&harr;'    => "\x{2194}",
              '&crarr;'   => "\x{21B5}",
              '&lArr;'    => "\x{21D0}",
              '&uArr;'    => "\x{21D1}",
              '&rArr;'    => "\x{21D2}",
              '&dArr;'    => "\x{21D3}",
              '&hArr;'    => "\x{21D4}",
              '&bull;'    => "\x{2022}",
              '&prime;'   => "\x{2032}",
              '&Prime;'   => "\x{2033}",
              '&oline;'   => "\x{203E}",
              '&frasl;'   => "\x{2044}",
              '&weierp;'  => "\x{2118}",
              '&image;'   => "\x{2111}",
              '&real;'    => "\x{211C}",
              '&trade;'   => "\x{2122}",
              '&euro;'    => "\x{20AC}",
              '&alefsym;' => "\x{2135}",
              '&spades;'  => "\x{2660}",
              '&clubs;'   => "\x{2663}",
              '&hearts;'  => "\x{2665}",
              '&diams;'   => "\x{2666}",
              '&OElig;'   => "\x{152}",
              '&oelig;'   => "\x{153}",
              '&Scaron;'  => "\x{160}",
              '&scaron;'  => "\x{161}",
              '&Yuml;'    => "\x{178}",
              '&fnof;'    => "\x{192}",
              '&ensp;'    => "\x{2002}",
              '&emsp;'    => "\x{2003}",
              '&thinsp;'  => "\x{2009}",
              '&zwnj;'    => "\x{200C}",
              '&zwj;'     => "\x{200D}",
              '&lrm;'     => "\x{200E}",
              '&rlm;'     => "\x{200F}",
              '&ndash;'   => "\x{2013}",
              '&mdash;'   => "\x{2014}",
              '&lsquo;'   => "\x{2018}",
              '&rsquo;'   => "\x{2019}",
              '&ldquo;'   => "\x{201C}",
              '&rdquo;'   => "\x{201D}",
              '&dagger;'  => "\x{2020}",
              '&Dagger;'  => "\x{2021}",
              '&hellip;'  => "\x{2026}",
              '&permil;'  => "\x{2030}",
              '&lsaquo;'  => "\x{2039}",
              '&rsaquo;'  => "\x{203A}",
              '&circ;'    => "\x{2C6}",
              '&tilde;'   => "\x{2DC}",
              );

###############
# convert XML-Escaped Chars
sub __xml_clean_values
{
  my $str=shift;
  $str=~s/(&\w+;)/exists($xmlchars{$1}) ? $xmlchars{$1} : int($1)>0 ? chr($1) : $1/gse;
  $str=Encode::decode('UTF-8', $str);
  return $str;
}

1;
